'use strict';

const express = require('express');
const bodyParser = require('body-parser');
const logger = require('./logger');
//const appmanager = require('./appmanager');
const http = require('http');
const https = require('https');
const async = require('async');
const fs = require("fs");
const request = require('request');
const path = require('path');
const Joi = require('joi');
const uuid = require('uuid-random');
const os = require('os');
const url = require('url');

const yaml_config = require('node-yaml-config');
logger.logger.info('***************************************** STARTING *****************************************');
logger.logger.info('Loading napi, make sure all DLL files are in place');
const nativeTest = require('./build/Release/native.node');
logger.logger.info('napi: DONE');

//###################################################
//                  Config
//###################################################
logger.logger.info('Loading config, make sure config.yml has valid structure');
var buildVersion = null;
var config = yaml_config.load(__dirname + '/config.yml');
logger.logger.info('config: DONE');

//###################################################
//                  Async calls
//###################################################

var currentRequests = {};
function setRequestProcessing(uuid, requestObj) {
    if (currentRequests[uuid] == null || currentRequests[uuid].state == null) {
        currentRequests[uuid] = { ts: new Date().getTime(), state: 'processing', req: requestObj };
        return;
    }
    if (currentRequests[uuid].state != 'success' && currentRequests[uuid].state != 'error' && currentRequests[uuid].state != 'cancelled')
        currentRequests[uuid].state = 'processing';
}
function updateCurrentResponseObj(uuid, responseObj) {
    if (currentRequests[uuid] == null)
        return;
    currentRequests[uuid].result = responseObj;
}
function setRequestSuccess(uuid, responseObj) {
    if (currentRequests[uuid] == null) {
        currentRequests[uuid] = { ts: new Date().getTime(), state: '' };
    }
    currentRequests[uuid].state = 'success';
    currentRequests[uuid].result = responseObj;
}
function setRequestError(uuid, message, responseObj) {
    if (currentRequests[uuid] == null) {
        currentRequests[uuid] = { ts: new Date().getTime(), state: '', message: '' };
    }
    currentRequests[uuid].state = 'error';
    currentRequests[uuid].message = message;
    currentRequests[uuid].result = responseObj;
}
function setRequestCancelled(uuid, responseObj) {
    if (currentRequests[uuid] == null || currentRequests[uuid].state == null) {
        currentRequests[uuid] = { ts: new Date().getTime(), state: 'cancelled' };
        return;
    }

    if (currentRequests[uuid].state != 'error' && currentRequests[uuid].state != 'success') {
        currentRequests[uuid].state = 'cancelled';
        currentRequests[uuid].result = responseObj;
    }
}

//###################################################
//                  Methods
//###################################################

const app = express();

app.use(bodyParser.json({ limit: '80mb' }));

app.use(function (req, res, next) {
    res.header("Access-Control-Allow-Origin", "*");
    res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
    next();
});

var api_definition = {
    "/v1/license/check": {
        method: 'get',
        func: (req, res) => { processLicenseCheck(req, res); },
        description: 'Pass ?json=true/false to change the output format',
        schema_resp: {
            "ok": "bool: true/false",
            "text": "string",
            "validTo": "string",
            "licenseTo": "string",
            "expired": "bool: true/false"
        }
    },
    "/v1/license/upgrade": {
        method: 'post',
        func: (req, res) => { processLicenseUpgrade(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "text": "string",
            "validTo": "string",
            "licenseTo": "string",
            "expired": "bool: true/false"
        }
    },
    "/v1/ingenico_status": {
        method: 'get',
        func: (req, res) => { processIngenicoStatus(req, res); },
        schema_resp: {
            "ok": "bool: true/false", "watermark" : "string", "error": "string", "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "transactionStatus": "number: (0x00 - 0x81) " +
                    "0x00 - Idle, " +
	                "0x01 - WaitCard, " +
                    "0x02 - WaitPIN, " +
                    "0x03 - WaitEMVApp, " +
                    "0x04 - WaitHost, " +
                    "0x05 - WaitSign, " +
                    "0x06 - WaitTrEnd, " +
                    "0x07 - WaitNoCard, " +
                    "0x08 - WaitBusy, " +
                    "0x09 - InProgress, " +
                    "0x0A - WaitCopy, " +
                    "0x0B - WaitAuthCode, " +
                    "0x0C - WaitAction, " +
                    "0x0D - BatchCompleted, " +
                    "0x0E - DCCCurrency, " +
                    "0x0F - CashBackAmount, " +
                    "0x10 - Used internaly, " +
                    "0x11 - Used internaly, " +
                    "0x12 - Transaction accepted, " +
                    "0x13 - Used internaly, " +
                    "0x14 - Waiting for amount, " +
                    "0x15 - Waiting for selection, " +
                    "0x16 - Asking ECR to print data, " +
                    "0x17 - Presenting ZenCard offer, " +
                    "0x18 - Waiting for fid, " +
                    "0x19 - Asking ECR to receive data, " +
                    "0x20 - Asking ECR to provide data, " +
                    "0x21 - Zencard continue, " +
                    "0x80 - Application in error state, " +
                    "0x81 - Reconciliation needed ",
                "sequenceNb": ""
            }
        },
    },
    "/v1/ingenico_general_info": {
        method: 'get',
        func: (req, res) => { processIngenicoGeneralInfo(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "terminalInfo": "string - comma separated data"
            }
        }
    },
    "/v1/ingenico_auto_transaction_process": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionAuto(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('purchase', 'preauthorization', 'completion', 'preauthorization_cancel', 'refund', 'reversal', 'giftcard_activation', 'giftcard_balance', 'giftcard_topup', 'reconciliation', 'connection_test', 'tms'),
            amount: Joi.number().required(),
            title: Joi.string(),//.regex(/^\s*\w+(?:[^\w,]+\w+)*[^,\w]*$/),
            /*
            ^ - start of string
            \s* - 0 or more whitespaces (or, use [^,\w]* to match 0 or more chars other than comma and word chars)
            \w+ - 1 or more word chars (letters, digits or _, if you do not want _, replace with [^\W_])
            (?:[^\w,]+\w+)* - zero or more repetitions of
            [^\w,]+ - 1 or more chars other than comma and word chars
            \w+ - 1 or more word chars
            [^,\w]* - 0 or more chars other than comma and word chars
            $ - end of string.
             */
            sequenceNb: Joi.string(),
            callback: Joi.object().keys({
                uri: Joi.string().required(),
                method: Joi.string().required().valid('POST', 'GET', 'PUT')
            }),
            posnetserver: Joi.object(),
            arguments: Joi.object().keys({
                transactionFlags: Joi.array().items(Joi.string().valid('TokenGeneration' /*0x01*/, 'ReferenceNumberGeneration'/*0x02*/, 'TransitTransactionIndicator'/*0x04*/)),     //DF01
                blikCode: Joi.number(),             //DF02 – BLIK code
                additionalPrintout: Joi.string(),   //DF03 – Additional printout data
                hints: Joi.string(),                //DF04 – "Hints"
                rfu: Joi.string(),                  //DF05 – RFU
                variableSymbol: Joi.string(),       //DF06 – Variable Symbol
                specificSymbol: Joi.string(),       //DF07 – Specific Symbol
                saleItemList: Joi.array().items(Joi.object().keys({   //7F01 – Sale Item List
                    itemId: Joi.string(),               //DF01 - Item ID 
                    productCode: Joi.string(),          //DF02 – Product Code 
                    amount: Joi.number(),               //DF03 - Amount(Total cost of item)
                    unitMeasure: Joi.string(),          //DF04 – Unit Measure
                    unitPrice: Joi.number(),            //DF05 – Unit Price(Price of single unit of item) 
                    quantity: Joi.number(),             //DF06 - Quantity 
                    taxCode: Joi.number()               //DF07 – Tax Code
                }))
            })
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "uuid": "string"
        },
        description: 'Use posnetserver structure when integration with PosnetServer is enabled (config.posnetserver.active=true). Sample JSON object: {"lines": [{ "na": "Towar 1", "il": 1.0, "vtp": "23,00", "pr": 2350 },{ "na": "Towar 2", "il": 1.0, "vtp": "23,00", "pr": 1150 }], "summary": { "to": 3500 } }. PosnetServer documentation can be found here: https://blog.bigdotsoftware.pl/posnet-server-wydruk-paragonu/'
    },
    "/v1/ingenico_auto_transaction_cancel/:uuid": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionCancel(req, res); },
        description: 'Method markes request as cancelled and releases COM port. Transaction still have to be manually cancelled from terminal\'s keyboard',
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "uuid": "string"
        }
    },
    "/v1/ingenico_auto_transaction_status/:uuid": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionStatus(req, res); },
        schema_resp: {
            "watermark": "string",
            "ts": "long",
            "state": "string: success, error, cancelled, processing",
            "req": "object",
            "result": "object"
        }
    },
    "/v1/ingenico_auto_transaction_statuses": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionStatuses(req, res); },
        schema_resp: {
            "watermark": "string",
            "{uuid}": {
                "ts": "long",
                "state": "string: success, error, cancelled, processing",
                "req": "object",
                "result": "object"
            }
        }
        //{"4ad8de25-f2ff-41d6-8783-86feef39dbc1":{"ts":1603906995024,"state":"cancelled","req":{"type":"purchase","amount":"1000","title":"Hello"},"result":{"ok":true,"error":"","terminal":{"returnCode":0,"transactionType":1,"transactionResultCode":7,"transactionResultCodeText":"interrupted","transactionNumber":-100,"transactionDetails":{"authorizationCode":"","serverMessage":"","merchantID":"454500000000000","maskedCardNumber":"","date":"2020.10.28","time":"18:42:06","cardsetName":"","readerUsed":"?","readerUsedText":"unknown","terminalIdentifier":"45450318","authorizationMessage":"000000","authorizationType":"?","authorizationTypeText":"unknown","transactionAmount":"","cardholderAuthorizationMethod":"?","cardholderAuthorizationMethodText":"no_authorization","atc":"","ac":"","aid":"","tvr":"","tsi":"","title":"","printingIndicator":"1","transactionCurrency":"PLN","terminalCurrency":"","exchangeRate":"","dcc1":"","dcc2":"","cashAmount":"","spendingAmount":"","dcc_currencyExp":"","additionalInformation":"DF0B0400000100"}}}},"bb9753cb-05fc-43dd-84a2-9af098bcc2b9":{"ts":1603907062507,"state":"processing","req":{"type":"purchase","amount":"1000","title":"Hello"},"result":{"ok":true,"error":"","terminal":{"returnCode":0,"transactionStatus":1,"transactionStatusText":"WaitCard","sequenceNb":"192"}}}}
    },
    "/v1/ingenico_transaction": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionStart(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('purchase', 'preauthorization', 'completion', 'preauthorization_cancel', 'refund', 'reversal', 'giftcard_activation', 'giftcard_balance', 'giftcard_topup', 'reconciliation', 'connection_test', 'tms'),
            amount: Joi.number().required(),
            title: Joi.string(),
            sequenceNb: Joi.string().required(),
            arguments: Joi.object().keys({
                transactionFlags: Joi.array().items(Joi.string().valid('TokenGeneration' /*0x01*/, 'ReferenceNumberGeneration'/*0x02*/, 'TransitTransactionIndicator'/*0x04*/)),     //DF01
                blikCode: Joi.number(),             //DF02 – BLIK code
                additionalPrintout: Joi.string(),   //DF03 – Additional printout data
                hints: Joi.string(),                //DF04 – "Hints"
                rfu: Joi.string(),                  //DF05 – RFU
                variableSymbol: Joi.string(),       //DF06 – Variable Symbol
                specificSymbol: Joi.string(),       //DF07 – Specific Symbol
                saleItemList: Joi.array().items(Joi.object().keys({   //7F01 – Sale Item List
                    itemId: Joi.string(),               //DF01 - Item ID 
                    productCode: Joi.string(),          //DF02 – Product Code 
                    amount: Joi.number(),               //DF03 - Amount(Total cost of item)
                    unitMeasure: Joi.string(),          //DF04 – Unit Measure
                    unitPrice: Joi.number(),            //DF05 – Unit Price(Price of single unit of item) 
                    quantity: Joi.number(),             //DF06 - Quantity 
                    taxCode: Joi.number()               //DF07 – Tax Code
                }) )
            })
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
        },
        description: "sequenceNb is mandatory but for completion and reversal types only. Set it to '-' for other types"
    },
    "/v1/ingenico_transaction_last/:transactionIndex": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionLast(req, res); }
    },
    "/v1/ingenico_transaction_end": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionEnd(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "transactionType": "number",
                "transactionResultCode": "number",
                "transactionResultCodeText": "string",
                "transactionNumber": "number",
                "transactionDetails": {
                    "authorizationCode": "string",
                    "serverMessage": "string",
                    "merchantID": "string: example '454500000000000'",
                    "maskedCardNumber": "string: example '************0211'",
                    "date": "string: example format 2021.01.25",
                    "time": "string: example format 12:59:32",
                    "cardsetName": "string: example 'VISA'",
                    "readerUsed": "string: example 'C' - possible values: B Contactless, M Manual, C Magstripe, P ICC, ? Unknown",
                    "readerUsedText": "string: possible values 'magstripe', 'contactless', 'manual', 'icc', 'unknown'",
                    "terminalIdentifier": "string: example '45450318'",
                    "authorizationMessage": "string: example '000180'",
                    "authorizationType": "string: possible values 1 - online, 3 – offline, 4 – referral",
                    "authorizationTypeText": "string: possible values 'online', 'offline', 'referral'",
                    "transactionAmount": "string",
                    "cardholderAuthorizationMethod": "string: possible values A – PIN, @ - Signature, B – PIN and Signature, ? – no authorization",
                    "cardholderAuthorizationMethodText": "string: possible values 'pin', 'signature', 'pin_signature', 'no authorization', 'unknown'",
                    "atc": "string",
                    "ac": "string",
                    "aid": "string",
                    "tvr": "string",
                    "tsi": "string",
                    "title": "string",
                    "printingIndicator": "string: example '1'",
                    "transactionCurrency": "string: example 'PLN'",
                    "terminalCurrency": "string",
                    "exchangeRate": "string",
                    "dcc1": "string",
                    "dcc2": "string",
                    "cashAmount": "string",
                    "spendingAmount": "string",
                    "dcc_currencyExp": "string",
                    "additionalInformation": "string: example 'DF060A534E204E722056697361DF0B0400000100' or 'DF0B0400000100' or 'DF060A534E204E722056697361DF0B0400000400DF1318343534353033313830303031323630313231323132383534'",
                    "additionalInformationObject": "object"
                },
                "error": "string"
            }
        }
    },
    "/v1/ingenico_batch_recon_data/:transactionIndex": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionReconData(req, res); }
    },
    "/v1/ingenico_batch_recon_info": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionReconInfo(req, res); },
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string",
            "terminal": {
                "returnCode": "number: (0,1,2,7) - 0 accepted, 1 refused, 2 No connection, 7 interrupted by the user",
                "error": "string",
                "results": "[] array"
            }
        }
    },
    "/v1/ingenico_transaction_data_input": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionDataInput(req, res); },
        schema: Joi.object().keys({
            text: Joi.string().required(),
            sequenceNb: Joi.string().required()
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string"
        }
    },
    "/v1/ingenico_transaction_send_key": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionSendKey(req, res); },
        schema: Joi.object().keys({
            keycode: Joi.string().required().valid('ok', 'back', 'cancel'),
            sequenceNb: Joi.string().required()
        }),
        schema_resp: {
            "ok": "bool: true/false",
            "watermark": "string",
            "error": "string"
        }
    },
    "/v1/ingenico_set_parameter": {
        method: 'post',
        func: (req, res) => { processIngenicoTerminalSetParameter(req, res); },
        schema: Joi.object().keys({
            type: Joi.string().required().valid('volatile', 'permanent'),
            parameters: Joi.array().required().items(Joi.object().keys({
                key: Joi.string().required(),
                value: Joi.string().required()
            }))
        })
    },
    "/v1/ingenico_set_standby": {
        method: 'post',
        func: (req, res) => { processIngenicoTerminalSetStandby(req, res); },
        schema: Joi.object().keys({
            timeoutSeconds: Joi.number().required()
        })
    },
    
    "/v1/ingenico_report_generate": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionReportGenerate(req, res); }
    },
    "/v1/ingenico_report_details/:cardset": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionGetSummary(req, res); }
    },
    "/v1/ingenico_report_summary": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionGetTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/v1/ingenico_analysis_generate": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionAnalysisGenerate(req, res); }
    },
    "/v1/ingenico_analysis_details/:transactionIndex": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionAnalysisDetails(req, res); }
    },
    "/v1/ingenico_analysis_summary": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionAnalysisTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/v1/ingenico_dcc_analysis_generate": {
        method: 'post',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisGenerate(req, res); }
    },
    "/v1/ingenico_dcc_analysis_details/:transactionIndex": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisDetails(req, res); }
    },
    "/v1/ingenico_dcc_analysis_summary": {
        method: 'get',
        func: (req, res) => { processIngenicoTransactionDCCAnalysisTotalSummary(req, res); }
    },

    //------------------------------------------------------------------------------------------------------------------
    "/": {
        method: 'get',
        func: (req, res) => { processRootRestApi(req, res); }
    }
}

function convertToSampleJSON(schema) {
    var result = {};
    if (schema['$_terms'].keys != null) {
        schema['$_terms'].keys.forEach(key => {
            //console.log('aaaaaaaaaaaaaaaaaa:' + key.key);

            if (key.schema.type == 'object') {
                //console.log('--------------------');
                //console.log(key.schema);
                var rrr = JSON.stringify(convertToSampleJSON(key.schema));
                //console.log('got --------------------' + rrr);
                result[key.key] = rrr;
            } else {
                result[key.key] = key.schema.type;
                if (key.schema['_valids'] != null && key.schema['_valids']['_values'] != null) {
                    //console.log(key.schema['_valids']['_values']);
                    result[key.key] += ': ';
                    const array = [...key.schema['_valids']['_values']];
                    result[key.key] += array.join(', ');
                }
            }
        });
    }
    return result;
}

function getProtocol(req) {
    var proto = req.connection.encrypted ? 'https' : 'http';
    proto = req.headers['x-forwarded-proto'] || proto;
    return proto.split(/\s*,\s*/)[0];
}

function GetWatermarkString() {

    var stat = CheckLicense();
    if (stat.ok) {
        var statinfo = readStatInfo(stat);
        return 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' ważna do ' + new Date(statinfo.validTo * 1000).toISOString();
    } else {
        return 'Wersja nielicencjonowana' + (config.license.file != '' ? ' lub przeznaczona dla innego produktu' : '') + '. W wersji nielicencjonowanej do kwoty na terminalu dodawane są losowe grosze';
    }
}

function initOtherOptions(other_options) {
    buildVersion = other_options.buildVersion;
}

function isFullDebugMode(req) {
    if (req == null) return false;
    if (req.query == null) return false;
    if (req.query.fulldebug == null) return false;
    if (req.query.fulldebug == "false") return false;

    if (req.query.fulldebug == "true") return true;

    return false;

    //return req.query.fulldebug != null ? (req.query.fulldebug ? true : false) : false;
}


function getCashRegisterNumber(req) {
    if (req == null) return 1;
    if (req.query == null) return 1;
    if (req.query.cashRegisterNumber == null) return 1;
    if (isNaN(req.query.cashRegisterNumber)) return 1;
    return req.query.cashRegisterNumber;
}

function isResetFlagEnabled(req) {
    if (req == null) return false;
    if (req.query == null) return false;

    if (req.query.reset == null) return false;
    if (req.query.reset == "false") return false;

    if (req.query.reset == "true") return true;

    return false;

    //return req.query.reset != null ? (req.query.reset ? true : false) : false;
}
function initApiDefinitions(def) {
    for (const [key, value] of Object.entries(def)) {
        if (value.method == 'get')
            app.get(key, value.func);
        else if (value.method == 'put')
            app.put(key, value.func);
        else if (value.method == 'post')
            app.post(key, value.func);
        else if (value.method == 'delete')
            app.delete(key, value.func);
        else
            logger.logger.error('Unrecognized method: ' + value.method);
    }
}

function initHiddenApiDefinitions() {
    app.get('/v1/tests/run', (req, res) => { processRunTests(req, res) } );
}

function processRootRestApi(req, res) {
    var restpoints = [
        { groupname: 'General information' },
        { type: 'GET', name: '' },
        { groupname: 'IngenicoServer methods (summary)' }
    ];

    for (const [key, value] of Object.entries(api_definition)) {
        if (key != "/" && key.startsWith("/auto_queue") == false )
            restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1) });
    }
    
    if (req.query.textdetails) {
        restpoints.push({ groupname: 'IngenicoServer methods (details)' });
        for (const [key, value] of Object.entries(api_definition)) {
            if (key != "/" && key.startsWith("/auto_queue") == false)
                restpoints.push({ type: value.method.toUpperCase(), name: key.substr(1), schema: value.schema, schema_resp: value.schema_resp, description: value.description });
        }
    }

    var restpointsstring = '';    
    restpoints.forEach(rp => {
        if (rp.groupname != null) {
            restpointsstring += '<br>\n<b>' + rp.groupname + '</b><br>\n';
        } else if (rp.ref != null) {
            restpointsstring += api_definition[rp.ref].method.toUpperCase().padEnd(6, ' ').replace(/ /g, '&nbsp;') + ' ' + getProtocol(req) + '://' + req.headers.host + '/' + rp.ref.substr(1) + '  <br>\n';
            if (api_definition[rp.ref].attributes != null && api_definition[rp.ref].attributes.length > 0) {
                api_definition[rp.ref].attributes.forEach(att => {
                    restpointsstring += '       '.replace(/ /g, '&nbsp;') + ' ?' + att.attribute + ' = ' + att.description + ' (default: ' + att.default + ')  <br>\n';
                });
            }
        } else {
            restpointsstring += rp.type.padEnd(6, ' ').replace(/ /g, '&nbsp;') + ' ' + getProtocol(req) + '://' + req.headers.host + '/' + rp.name + '  <br>\n';

            if (rp.schema)
                restpointsstring += 'Request:<br>\n' + prettyJson(convertToSampleJSON(rp.schema)).replace(/\t/g, '<br>') + '  <br>\n';

            if (rp.schema_resp)
                restpointsstring += 'Response:<br>\n' + prettyJson(rp.schema_resp).replace(/\t/g, '<br>') + '  <br>\n';

            if (rp.description)
                restpointsstring += 'Description:<br>\n' + (rp.description) + '  <br>\n';

            if (rp.schema || rp.schema_resp)
                restpointsstring += '  <hr>\n';
            //restpointsstring += '  <hr>\n';
        }
    });

    if (!req.query.textdetails) {
        //restpointsstring += '<br>\n<a href="/?textdetails=true">Show</a><br>\n';  //be zsensu dluzej to pokazywac skro calal dokumentacja i tak jest w swagger
    }

    res.send(
        '<tt>\n' +
        'Ingenico Server v.' + buildVersion + ' © 2021 BigDotSoftware (bigdotsoftware@bigdotsoftware.pl). <br>\n' +
        'How to start: https://blog.bigdotsoftware.pl/ingenico-server-pierwsze-uzycie/ <br>\n<br>\n' +
        'Swagger YAML: &nbsp; <a href="https://github.com/bigdotsoftware/ingenicoserver/blob/main/swagger/swagger-' + buildVersion + '.yaml">https://github.com/bigdotsoftware/ingenicoserver/blob/main/swagger/swagger-' + buildVersion + '.yaml</a> <br>\n<br>\n' +
        'Swagger Editor: <a href="https://editor.swagger.io/?url=https://download2.bigdotsoftware.pl/github_ingenico_swagger/swagger-' + buildVersion + '.yaml">Open in Editor</a> <br>\n<br>\n' +
        'RESTful API: <br></tt><tt>\n' +
        restpointsstring + 
        '</tt>'
    );
}

function replaceAll(string, search, replace) {
    return string.split(search).join(replace);
}

function ConnectionBegin(req, finalcb) {

    var fulldebug = isFullDebugMode(req);
    //FetchRemoteConfig(req, res, (err) => {
    //    if (err) {
    //        logger.logger.warn('FetchRemoteConfig error: ' + err);
    //        finalcb(true);  //return error
    //        return;
    //    }

    //    ValidateScurityToken(req, res, (result, err) => {
    //        if (err) {          //communication error
    //            finalcb(true);  //return error
    //        }
    //        else if (result == false) {     //validation didnt pass
    //            logger.logger.warn('Secutity token doesn\'t match');
    //            tools.resp_error(res, 'Secutity token doesn\'t match');
    //            finalcb(true);  //return error
    //        }
    //        else {
                if (nativeTest.PosnetIsConnected().ok) {
                    finalcb({ ok: false, error: 'Already connected - check connect sequence or another transaction is in progress' });
                    //already connected
                    //finalcb({ ok: true, error: 'Already connected' });
                } else {
                    if (fulldebug) console.log('FULLDEBUG: Preparing connection to the printer');
                    var rc = ConnectionBeginProcess(req);
                    finalcb(rc);
                }
    //        }
     //   });
    //});

}


function ConnectionBeginProcess(req) {
    var fulldebug = isFullDebugMode(req);
    var stat = null;
    if (config.ingenico.connection.type == 'rs232') {
        if (fulldebug) console.log('FULLDEBUG: Opening COM port ' + config.ingenico.connection.rs232.params);
        const disable_ioctl = config.ingenico.connection.rs232.disable_ioctl != null ? config.ingenico.connection.rs232.disable_ioctl : false;
        stat = nativeTest.PosnetConnectRS232(config.ingenico.connection.rs232.params, fulldebug ? "1" : "0", disable_ioctl ? "1" : "0");
        if (fulldebug) console.log('FULLDEBUG: Opening COM port result' + JSON.stringify(stat));
    } else if (config.ingenico.connection.type == 'eth') {
        if (fulldebug) console.log('FULLDEBUG: Opening Socket ' + config.ingenico.connection.eth.ip + ':' + config.ingenico.connection.eth.port);
        stat = nativeTest.PosnetConnectEth(config.ingenico.connection.eth.ip, config.ingenico.connection.eth.port, fulldebug ? "1" : "0");
    }

    if (stat == null || stat.ok == false) {
        logger.logger.error(JSON.stringify(stat));
        //tools.resp_error(res, stat.error);
        return (stat == null ? { ok: false, error: 'Unknown error' } : stat);
    } else {
        return stat;
    }
}

function ConnectionEnd() {
    var stat = nativeTest.PosnetDisconnect();
    if (!stat.ok) {
        logger.logger.error(JSON.stringify(stat));
        //tools.resp_error(res, stat.error);
    }
    return stat;
}

function readStatInfo(stat) {
    var licenseTo = '';
    var validTo = 0;
    var expired = true;
    stat.results.forEach(result => {
        if (result.licenseTo != null && result.licenseTo != '')
            licenseTo = result.licenseTo;
        if (result.validTo != null && result.validTo != '')
            validTo = result.validTo;
        if (result.expired != null && result.expired != '')
            expired = result.expired == 'false' ? false : true;
    });
    return {
        licenseTo: licenseTo,
        validTo: validTo,
        expired: expired
    };
}

function obtainNewLicenseFile(uniqueLicenseId) {
    var fullurl = 'https://license.bigdotsoftware.pl/obtain/' + uniqueLicenseId;
    logger.logger.debug(fullurl);
    request({
        url: fullurl,
        method: 'POST',
        rejectUnauthorized: false,
        strictSSL: false,
        encoding: null /* body as buffer*/,
        json: {
            'action': 'renew'
        }
    }, (err, internalres, body) => {
            logger.logger.info('statusCode:', internalres && internalres.statusCode);
        if (err)
            logger.logger.info('cannot execute license endpoint: ' + err);
        else if (internalres && internalres.statusCode != 200)
            logger.logger.error('cannot execute license endpoint, statusCode: ' + internalres.statusCode);
        else {
            logger.logger.info('license endpoint executed ok');
            var fileDestination = config.license.file + '.' + new Date().getTime() + '.backup';
            logger.logger.info('license endpoint copy as ' + fileDestination);

            fs.copyFile(config.license.file, fileDestination, (err) => {
                if (err) {
                    logger.logger.error('error ' + err);
                } else {
                    logger.logger.info('copied as ' + fileDestination);
                    fs.writeFileSync(config.license.file + '.new', body, 'binary');

                    logger.logger.info('validating ' + config.license.file + '.new');
                    var stat = nativeTest.ValidateLicenseFile({ file: config.license.file + '.new', skipcache: 1 });
                    logger.logger.info('validated' + JSON.stringify(stat));
                    if (stat.ok) {
                        logger.logger.info('downloaded file is valid license file');
                        var statinfo = readStatInfo(stat);
                        logger.logger.debug(statinfo);
                        if (!statinfo.expired) {
                            logger.logger.info('replacing license file....');
                            fs.copyFile(config.license.file + '.new', config.license.file, (err) => {
                                if (err)
                                    logger.logger.error('error replacing license file: ' + err);
                                else
                                    logger.logger.info('license file replaced');
                            });
                        } else {
                            logger.logger.warn('... but expired');
                        }
                    } else {
                        logger.logger.error('downloaded file is not valid license file');
                    }
                }
            });
        }
    });
}

function processLicenseUpgrade(req, res) {
    setTimeout(function () {

        if (config.license.file != null && config.license.file != '') {
            logger.logger.info('license file: ' + config.license.file);
            var stat = nativeTest.ValidateLicenseFile({ file: config.license.file });
            logger.logger.info(JSON.stringify(stat));
            if (stat.results != null) {
                var uniqueLicenseId = '';
                stat.results.forEach(result => {
                    if (result.uniqueLicenseId != null && result.uniqueLicenseId != '') {
                        uniqueLicenseId = result.uniqueLicenseId;
                    }
                });
                if (uniqueLicenseId != '')
                    obtainNewLicenseFile(uniqueLicenseId);
                else
                    logger.logger.error('Cannot determine unique license ID');

                if (stat.ok) {
                    var statinfo = readStatInfo(stat);
                    return sendResponse(req, res, {
                        ok: true,
                        text: 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' ważna do ' + new Date(statinfo.validTo * 1000).toISOString(),
                        validTo: statinfo.validTo,
                        licenseTo: statinfo.licenseTo,
                        expired: statinfo.expired,
                        requested: true,
                        requestedtext: 'Requested for new license file. Check logs for more details'
                    });
                } else {
                    return sendErrorResponse(req, res, 'Wersja nielicencjonowana');
                }
            }
        } else {
            return sendErrorResponse(req, res, 'Wersja nielicencjonowana');
        }

    }, 500);
}

function processLicenseCheck(req, res) {
    setTimeout(function () {
        var stat = CheckLicense();
        if (stat.ok) {
            var statinfo = readStatInfo(stat);

            if (req.query.json)
                return sendResponse(req, res, {
                    ok: true,
                    text: 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' ważna do ' + new Date(statinfo.validTo * 1000).toISOString(),
                    validTo: statinfo.validTo,
                    licenseTo: statinfo.licenseTo,
                    expired: statinfo.expired
                });
            else
                return sendResponse(req, res, 'Wersja licencjonowana dla ' + statinfo.licenseTo + ' ważna do ' + new Date(statinfo.validTo * 1000).toISOString());
        } else {
            if (req.query.json)
                return sendErrorResponse(req, res, {
                    ok: false,
                    text: 'Wersja nielicencjonowana',
                    expired: true
                });
            else
                return sendErrorResponse(req, res, 'Wersja nielicencjonowana');
        }
    }, 500);
}

function processRunTests(req, res) {
    var fulldebug = isFullDebugMode(req);
    const msg0 = nativeTest.RunTests(fulldebug ? "1" : "0");
    return sendResponse(req, res, msg0);
}

function processIngenicoStatus(req, res) {

    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        const msg0 = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber } );

        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });


    /*

    var fulldebug = true;
    var stat = nativeTest.GeneralInformation(fulldebug ? "1" : "0");
    if (!stat.ok) {
        logger.logger.error(JSON.stringify(stat));
        return sendErrorResponse(req, res, stat.error);
    }
    //finalresults.ok = true;
    return sendResponse(req, res, {});
    */
    
}

function processIngenicoGeneralInfo(req, res) {

    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        const msg0 = nativeTest.GetTermGeneralInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber } );

        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function verifyTermResponse(req, res, result, uuid /*optional*/) {
    if (result.terminal.returnCode != 0) {
        if (res != null )
            sendErrorResponse(req, res, result);

        var closeresult = ConnectionEnd();
        /*if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);*/
        if (uuid != null)
            setRequestError(uuid, result.error, result);

        return false;
    } else {
        return true;
    }
}

function processIngenicoTransactionReconData(req, res) {
    var fulldebug = isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconData = nativeTest.TransactionReconData(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconData)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trReconData);
    });
}

function processIngenicoTransactionLast(req, res) {
    var fulldebug = isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconData = nativeTest.TransactionLastData(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconData)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trReconData);
    });
}


function processIngenicoTransactionReconInfo(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconInfo = nativeTest.TransactionReconInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trReconInfo);
    });
}

function processIngenicoTransactionDataInput(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const trReconInfo = nativeTest.TransactionDataInput(fulldebug ? "1" : "0", { text: req.body.text, sequenceNb: req.body.sequenceNb, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trReconInfo);
    });
}

function processIngenicoTransactionSendKey(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        var keycode = '';
        if (req.body.keycode == 'ok')
            keycode = "\\n"
        else if (req.body.keycode == 'back')
            keycode = "\\b"
        else if (req.body.keycode == 'cancel')
            keycode = "\\e"

        const trReconInfo = nativeTest.TransactionDataInput(fulldebug ? "1" : "0", { text: keycode, sequenceNb: req.body.sequenceNb, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trReconInfo);
    });
}


function processIngenicoTerminalSetParameter(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TerminalSetParameter(fulldebug ? "1" : "0", {
            parameters: req.body.parameters,
            typeOfParameter: (req.body.type == 'permanent' ? 1 : 0),    //('volatile', 'permanent')
            cashRegisterNumber: cashRegisterNumber
        });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}


function processIngenicoTerminalSetStandby(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TerminalSetStandby(fulldebug ? "1" : "0", {
            timeoutSeconds: req.body.timeoutSeconds,
            cashRegisterNumber: cashRegisterNumber
        });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}



//-----------------------------------------------------------------------------
function processIngenicoTransactionReportGenerate(req, res) {
    var fulldebug = isFullDebugMode(req);
    var reset = isResetFlagEnabled(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGenerate(fulldebug ? "1" : "0", { reset: (reset ? "1" : "0"), cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionGetSummary(req, res) {
    var fulldebug = isFullDebugMode(req);
    const cardset = req.params.cardset;
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGetSummary(fulldebug ? "1" : "0", { cardset: cardset, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionGetTotalSummary(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionReportGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

//-----------------------------------------------------------------------------
function processIngenicoTransactionAnalysisGenerate(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGenerate(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionAnalysisDetails(req, res) {
    var fulldebug = isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGetDetails(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionAnalysisTotalSummary(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionAnalysisGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

//-----------------------------------------------------------------------------
function processIngenicoTransactionDCCAnalysisGenerate(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGenerate(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionDCCAnalysisDetails(req, res) {
    var fulldebug = isFullDebugMode(req);
    const transactionIndex = req.params.transactionIndex;
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGetDetails(fulldebug ? "1" : "0", { transactionIndex: transactionIndex, cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function processIngenicoTransactionDCCAnalysisTotalSummary(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        const msg0 = nativeTest.TransactionDCCAnalysisGetTotalSummary(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, msg0)) return;

        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, msg0);
    });
}

function ClearBatchCompleted(req, res, fulldebug) {
    var cashRegisterNumber = getCashRegisterNumber(req);

    for (var i = 0; i < 100; i++) {
        const trReconInfo = nativeTest.TransactionReconInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trReconInfo)) return false;

        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus)) return false;

        if (trStatus.terminal.transactionStatus == 0)
            break;
    }
    return true;
}

function processIngenicoTransactionEnd(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trEnd)) return;
        logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));
        
        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, trEnd);
    });
}

function sleep(ms) {
    return new Promise((resolve) => {
        setTimeout(resolve, ms);
    });
}

function getSchemaForRequest(req) {
    //console.log(req);
    let parsedUrl = url.parse(req.url);
    return api_definition[parsedUrl.pathname] != null ? api_definition[parsedUrl.pathname].schema : null;
}

function validateSchema(req, res) {
    var schema = getSchemaForRequest(req);
    if (schema != null) {
        var test = schema.validate(req.body);
        if (test.error != null) {
            logger.logger.warn('Payload validation error:' + JSON.stringify(test));
            sendErrorResponse(req, res, test.error);
            return false;
        } else {
            return true;
        }
    } else {
        return true;    //no validation - everythign is fine :-)
    }
}

function processIngenicoTransactionCancel(req, res) {
    //var fulldebug = isFullDebugMode(req);
    var uuid = req.params.uuid;

    if (currentRequests[uuid] != null) {
        logger.logger.info(`Marking request ${uuid} cancelled`);
        setRequestCancelled(uuid, { ok: false });

        return sendResponse(req, res, currentRequests[uuid]);
    } else {
        return sendErrorResponse(req, res, 'Wrong UUID');
    }
}

function processIngenicoTransactionStatus(req, res) {
    if (currentRequests[req.params.uuid] != null)
        return sendResponse(req, res, currentRequests[req.params.uuid]);
    else
        return sendErrorResponse(req, res, 'Wrong UUID');

}
function processIngenicoTransactionStatuses(req, res) {
    return sendResponse(req, res, currentRequests);
}
function processIngenicoTransactionAuto(req, res) {
    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, async (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        const transactionUUID = uuid();
        setRequestProcessing(transactionUUID, req.body);

        //------------------------
        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        if (trStatus.terminal.transactionStatus == 129 /* 0x81 TRANSACTION_STATUS_RECONCILIATION_NEEDED*/) {
            const trReconciliation = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
                type: 'reconciliation',
                amount: 0,
                title: '',
                sequenceNb: '',
                arguments: {},
                cashRegisterNumber: cashRegisterNumber,
                file: GetLicenseFile()
            });
            logger.logger.info('trReconciliation done: ' + JSON.stringify(trReconciliation));
            if (!verifyTermResponse(req, res, trReconciliation, transactionUUID)) return;

            trStatus = await startListeningForWaitTrEnd(req, res, fulldebug, transactionUUID);

            var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trEnd, transactionUUID)) return;
            logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

            trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
            logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
        }

        if (trStatus.terminal.transactionStatus == 13 /* 0x0D TRANSACTION_STATUS_BATCH_COMPLETED*/) {
            var rc = ClearBatchCompleted(req, res, fulldebug);
            if (!rc) {
                ConnectionEnd();
                return;
            }
        }


        if (trStatus.terminal.transactionStatus == 1 /* 0x01 TRANSACTION_STATUS_WAIT_CARD*/) {
            logger.logger.warn("Cannot interrupt transaction in WAIT_CARD state");

            ConnectionEnd();
            sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Przyłuż kartę lub anuluj na terminalu rozpoczętą transakcję' });
            return;
        }

        if (trStatus.terminal.transactionStatus == 6 /* 0x06 TRANSACTION_STATUS_WAIT_TREND*/) {

            logger.logger.info("Interrupting current transaction in WAIT_TREND state");

            var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trEnd, transactionUUID)) return;
            logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

            trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
            if (!verifyTermResponse(req, res, trStatus, transactionUUID)) return;
            logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
        }

        //------------------------
        //var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trEnd)) return;
        //logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        //------------------------
        //trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trStatus)) return;
        //logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        //------------------------
        const trBegin = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
            type: req.body.type/* 'purchase'*/,   //purchase=1, preauthorization=4, completion=5, preauthorization_cancel=82, refund=6, reversal=10, giftcard_activation=42, giftcard_balance=43, giftcard_topup=44, reconciliation=31, connection_test=33, tms=40
            amount: req.body.amount != null ? req.body.amount : 0 /*1000*/,
            title: req.body.title != null ? req.body.title : '', //'Purchase 1234'
            sequenceNb: req.body.sequenceNb != null ? req.body.sequenceNb : '',
            arguments: req.body.arguments != null ? req.body.arguments : {},
            cashRegisterNumber: cashRegisterNumber,
            file: GetLicenseFile()
        });
        logger.logger.info('trBegin done: ' + JSON.stringify(trBegin));
        if (!verifyTermResponse(req, res, trBegin, transactionUUID)) return;

        //------------------------
        if (!trBegin.ok)
            sendErrorResponse(req, res, trBegin);
        else
            sendResponse(req, res, { ok: true, uuid: transactionUUID });

        trStatus = await startListeningForWaitTrEnd(req, null/*res - pas null to avoid sending response*/, fulldebug, transactionUUID);

        if (trStatus.terminal.transactionStatus != 6 /*WaitTrEnd*/) {
            logger.logger.info('Interrupting due to wrong status: ' + JSON.stringify(trStatus));
            ConnectionEnd();
            return;
        }

        var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, null /*res - pas null to avoid sending response*/, trEnd, transactionUUID)) return;
        logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        if (trEnd.terminal.transactionResultCode ==0 )
            setRequestSuccess(transactionUUID, trEnd);
        else if (trEnd.terminal.transactionResultCode == 1 || trEnd.terminal.transactionResultCode == 7)
            setRequestCancelled(transactionUUID, trEnd);
        else
            setRequestError(transactionUUID, "", trEnd);

        if (req.body.callback) {
            notifyBackendAboutTransactionResult(trEnd, req.body.callback);
        }

        if (config.posnetserver.active) {
            if (req.body.posnetserver) {
                if (trEnd.terminal.transactionResultCode == 0) {
                    sendToPosnetServer(req.body.posnetserver);
                } else {
                    logger.logger.warn(`Integracja z PosnetServer jest włączona, transakcja na terminalu nie zakonczyla sie sukcesem - paragon nie bedzie drukowany`);
                }
            } else {
                logger.logger.warn(`Integracja z PosnetServer jest włączona, ale request nie zawiera struktury 'posnetserver' którą można przekazać do serwisu.`);
            }
        }


        var result = ConnectionEnd();
        if (!result.ok) {
            logger.logger.error('cannot end: ' + JSON.stringify(result));
        }
    });
}

function sendToPosnetServer(posnetserverObj) {
    logger.logger.warn(`Integracja z PosnetServer : ` + JSON.stringify(posnetserverObj) );
    ExecuteUrl(config.posnetserver.url + '/paragon', 'POST', posnetserverObj );
}

function ExecuteUrl(fullurl, resultsurlmethod, body) {

    const tt_start = new Date().getTime();

    var apiMethodWrapper = function (callback) {
        request({
            url: fullurl,
            method: resultsurlmethod,
            rejectUnauthorized: false,
            strictSSL: false,
            json: body
        }, (err, internalres, body) => {
            if (err) {
                logger.logger.error('Cannot execute url: ' + fullurl + ' : ' + err);
                callback(true);//retry
            } else {
                logger.logger.debug(`URL ${fullurl} executed OK`);
                callback(false, {});//ok, no need to retry
            }
        });
    };

    async.retry({ times: config.ingenico.retry.times, interval: config.ingenico.retry.interval }, apiMethodWrapper, function (err, responseObj) {

        logger.logger.debug(`ExecuteUrl is Done`);
        const took = new Date().getTime() - tt_start;
        if (err) {
            logger.logger.error(`ExecuteUrl failed in ${took} ms : ${err}`);
        } else {
            logger.logger.info(`ExecuteUrl succeeded ${took} ms`);
        }

    });

}

function notifyBackendAboutTransactionResult(trEnd, callbackObj) {
    ExecuteUrl(callbackObj.uri, callbackObj.method, JSON.stringify(trEnd));
}

async function startListeningForWaitTrEnd(req, res, fulldebug, uuid) {

    var cashRegisterNumber = getCashRegisterNumber(req);

    var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
    updateCurrentResponseObj(uuid, trStatus);
    if (!verifyTermResponse(req, res, trStatus)) return trStatus;
    logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));
    

    while (trStatus.terminal.transactionStatus != 6 /*WaitTrEnd*/) {
        trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        updateCurrentResponseObj(uuid, trStatus);
        if (!verifyTermResponse(req, res, trStatus)) return trStatus;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));


        if (currentRequests[uuid] != null && currentRequests[uuid].state == 'cancelled') {
            logger.logger.info(`Received cancelled flag for ${uuid}`);
            return trStatus;
        }

        await sleep(1000);
    }

    return trStatus;
}

function processIngenicoTransactionStart(req, res) {

    var fulldebug = isFullDebugMode(req);
    var cashRegisterNumber = getCashRegisterNumber(req);

    if (!validateSchema(req, res))
        return;

    ConnectionBegin(req, (err) => {
        if (!err.ok) {
            sendErrorResponse(req, res, err);
            return;
        }

        //------------------------
        var trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        if (!verifyTermResponse(req, res, trStatus)) return;
        logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        if (trStatus.terminal.transactionStatus == 13 /* 0x0D TRANSACTION_STATUS_BATCH_COMPLETED*/) {
            var rc = ClearBatchCompleted(req, res, fulldebug);
            if (!rc) {
                ConnectionEnd();
                return;
            }
        }
        

        if (trStatus.terminal.transactionStatus == 1 /* 0x01 TRANSACTION_STATUS_WAIT_CARD*/) {
            ConnectionEnd();
            sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Przyłuż kartę lub zamknij rozpoczętą transakcję' });
            return;
        }
        if (trStatus.terminal.transactionStatus == 6 /* 0x06 TRANSACTION_STATUS_WAIT_TREND*/) {
            ConnectionEnd();
            sendErrorResponse(req, res, { ok: false, error: trStatus.terminal.transactionStatusText + ' : Zamknij niedokończoną transakcję'});
            return;
        }

        //------------------------
        //var trEnd = nativeTest.TransactionEnd(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trEnd)) return;
        //logger.logger.info('trEnd done: ' + JSON.stringify(trEnd));

        //------------------------
        //trStatus = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
        //if (!verifyTermResponse(req, res, trStatus)) return;
        //logger.logger.info('trStatus done: ' + JSON.stringify(trStatus));

        //------------------------
        logger.logger.info('TransactionBegin - preparing');
        const trBegin = nativeTest.TransactionBegin(fulldebug ? "1" : "0", {
            type: req.body.type/* 'purchase'*/,   //purchase=1, preauthorization=4, completion=5, preauthorization_cancel=82, refund=6, reversal=10, giftcard_activation=42, giftcard_balance=43, giftcard_topup=44, reconciliation=31, connection_test=33, tms=40
            amount: req.body.amount != null ? req.body.amount : 0 /*1000*/,
            title: req.body.title != null ? req.body.title : '', //'Purchase 1234'
            sequenceNb: req.body.sequenceNb != null ? req.body.sequenceNb : '',
            arguments: req.body.arguments != null ? req.body.arguments : {},
            cashRegisterNumber: cashRegisterNumber,
            file: GetLicenseFile()
        });
        logger.logger.info('trBegin done: ' + JSON.stringify(trBegin));
        if (!verifyTermResponse(req, res, trBegin)) return;


        //------------------------
        var result = ConnectionEnd();
        if (!result.ok)
            return sendErrorResponse(req, res, result);
        else
            return sendResponse(req, res, {ok: true});
    });
}


function isPrettyOutput(req) {
    return req.query != null && req.query.pretty;
}
function prettyJson(obj) {
    return '<tt>' + JSON.stringify(obj, null, 2).replace(/\n/g, "<br/>").replace(/ /g, "&nbsp;") + '</tt>';
}

function sendErrorResponse(req, res, resultobj) {
    var watermark = GetWatermarkString();
    if (resultobj === Object(resultobj)) {
        resultobj['watermark'] = watermark;
    }

    if (isPrettyOutput(req))
        return res.send(prettyJson(resultobj /*{
            ok: false,
            terminal: resultobj
        }*/));
    else
        return res.json(resultobj /*{
            ok: false,
            terminal: resultobj
        }*/);
}

function sendResponse(req, res, resultobj) {
    var watermark = GetWatermarkString();
    if (resultobj === Object(resultobj)) {
        resultobj['watermark'] = watermark;
    }

    if (isPrettyOutput(req))
        return res.send(prettyJson(resultobj/*{
            ok: true,
            terminal: resultobj
        }*/));
    else
        return res.json(resultobj /*{
            ok: true,
            terminal: resultobj
        }*/);
}

function initNative() {

    //var useddriver = config.posnet.driver != null ? config.posnet.driver : 'posnet';
    var usedcommanddelay = config.ingenico.commanddelay != null ? config.ingenico.commanddelay : 200;
    var usedwaitdelay = config.ingenico.waitdelay != null ? config.ingenico.waitdelay : 100;
    var usedcommandresponsetimeout = config.ingenico.commandresponsetimeout != null ? config.ingenico.commandresponsetimeout : 10000;
    var responsewaitdelay = config.ingenico.responsewaitdelay != null ? config.ingenico.responsewaitdelay : 100;

    logger.logger.info('Initializing ingenico module');
    nativeTest.InitializeModule({
        // loggerterminal: config.logger.terminal,
        loggernative: config.logger.native,
        licensefile: GetLicenseFile(),
        //posnetdriver: useddriver,
        commanddelay: usedcommanddelay,
        waitdelay: usedwaitdelay,
        responsewaitdelay: responsewaitdelay,
        commandresponsetimeout: usedcommandresponsetimeout
    });
}

function initScheduleQueueCleaner() {

    logger.logger.info(`Scheduling queue cleaner with retention of ${config.queue.retention} ms`);
    setInterval(function () {

        const currentts = new Date().getTime();
        const retention = config.queue.retention;
        var newcurrentRequests = {};

        for (const [key, value] of Object.entries(currentRequests)) {

            var retention_offset = value.ts + retention;
            if (retention_offset < currentts && value.state != 'processing') {    //jelsi minelo 5 minut oraz inprogress jets na false
                var objtext = JSON.stringify(value);
                logger.logger.debug(`Removing expired ${key} from the queue -> ${objtext}`);
            } else {
                newcurrentRequests[key] = value;
            }
        }

        currentRequests = newcurrentRequests;

    }, 2000);
}

function initPinger() {
    if (config.ingenico.ping.active) {
        var cashRegisterNumber = config.ingenico.ping.cashRegisterNumber;
        if (cashRegisterNumber == null)
            cashRegisterNumber = 1;
        logger.logger.info(`Scheduling pinger with interval ${config.ingenico.ping.interval} ms on cashRegisterNumber=${cashRegisterNumber}`);
        setInterval(function () {

            var fulldebug = false;

            ConnectionBegin(null, (err) => {
                if (!err.ok) {
                    logger.logger.error('PING|Port open error: ' + JSON.stringify(err));
                    return;
                }

                const msg0 = nativeTest.GetTermInfo(fulldebug ? "1" : "0", { cashRegisterNumber: cashRegisterNumber });
                var txt = msg0.terminal.transactionStatusText != null ? msg0.terminal.transactionStatusText : '';
                logger.logger.info(`PING|status: ${msg0.terminal.transactionStatus} with text: '${txt}'`);

                var result = ConnectionEnd();
                if (!result.ok)
                    logger.logger.error('PING|Port close error: '+ JSON.stringify(result) );
            });

        }, config.ingenico.ping.interval);
    } else {
        logger.logger.info(`Pinger is disabled in configuration. To enable it set ingenico.ping.active=true`);
    }
}



function initLicense() {
    logger.logger.info(`Checking license file ${GetLicenseFile()} .....`);
    var stat = CheckLicense();
    return stat;
}
function init(http_options, https_options, other_options) {

    initApiDefinitions(api_definition);
    initHiddenApiDefinitions();
    initOtherOptions(other_options);
    initScheduleQueueCleaner();

    logger.logger.info(`Trying to open HTTP on port: ${http_options.port}`);
    try {
        var server = http.createServer(app);
        server.on('error', function (e) {
            logger.logger.error(`Cannot start HTTP at http://localhost: ${http_options.port}, please check if there no another process listening on the same port. Reason: ${e}`);
        });
        server.listen(http_options.port);
        logger.logger.info('HTTP  at http://localhost:' + http_options.port)
    } catch (e) {
        logger.logger.error(`Cannot start HTTP at http://localhost:' ${http_options.port}, please check if there no another process listening on the same port. Reason: ${e}`);
    }


    if (https_options.active) {
        logger.logger.info(`Loading SSL certs...: ${https_options.sslcertificates.key}`);
        var options = null;
        try {
            options = {
                key: fs.readFileSync(https_options.sslcertificates.key),
                cert: fs.readFileSync(https_options.sslcertificates.crt)
            };
        } catch (e) {
            logger.logger.error(`Cannot load certificates ${https_options.sslcertificates.key} and ${https_options.sslcertificates.crt}. Reason: ${e}`);
        }

        if (options != null) {
            logger.logger.info(`Trying to open HTTPS on port: ${https_options.port}`);
            try {
                https.createServer(options, app).listen(https_options.port);
                logger.logger.info(`HTTPS at https://localhost: ${https_options.port}`);
            } catch (e) {
                logger.logger.error(`Cannot start HTTPS at https://localhost: ${https_options.port}, please check your certificate under the ./cert directory. Reason: ${e}`);
            }
        }
    } else {
        logger.logger.info('HTTPS is disabled in configuration. To enable set https.active=true in config.yml');
    }

    initNative();
    initPinger();
    initLicense();
}

function GetLicenseFile() {
    return config.license.file;
}

function CheckLicense() {
    var stat = {ok: false};
    if (config.license.file == null || config.license.file == '') {
        logger.logger.warn('No license file provided. Amount on terminal in unregistered version is randomly increased by PLN grosze.');
        logger.logger.warn('Brak pliku licencji. W wersji nielicencjonowanej do kwoty na terminalu dodawane są losowe grosze.');
    } else {
        //if (config.license.url != null && config.license.url != '') {
            // TODO
        //}
        if (config.license.file != null && config.license.file != '') {
            logger.logger.info('license file: ' + config.license.file);
            stat = nativeTest.ValidateLicenseFile({ file: config.license.file });
            logger.logger.info(JSON.stringify(stat));
            if (!stat.ok) {
                logger.logger.warn('Wrong license file provided. Amount on terminal in unregistered version is randomly increased by PLN grosze.');
                logger.logger.warn('Błędny plik licencji. W wersji nielicencjonowanej do kwoty na terminalu dodawane są losowe grosze.');
            }

            if (stat.results != null) {
                var uniqueLicenseId = '';
                var days = 999;
                stat.results.forEach(result => {
                    if (result.validTo != null && result.validTo != '') {
                        //logger.logger.debug(parseInt(result.validTo));
                        var seconds = parseInt(result.validTo) - new Date().getTime() / 1000;
                        days = seconds / 60 / 60 / 24;
                        if (days < 5)
                            logger.logger.error(days + ' days left');
                        else if (days < 10)
                            logger.logger.warn(days + ' days left');
                        else
                            logger.logger.info(days + ' days left');
                    }

                    if (result.uniqueLicenseId != null && result.uniqueLicenseId != '') {
                        uniqueLicenseId = result.uniqueLicenseId;
                    }
                });

                //przedstaw sie jakiej licencji uzywasz i gdzie

                var ifaces = os.networkInterfaces();
                var ifacesarray = [];
                Object.keys(ifaces).forEach(function (ifname) {
                    var alias = 0;

                    ifaces[ifname].forEach(function (iface) {
                        if ('IPv4' !== iface.family || iface.internal !== false) {
                            // skip over internal (i.e. 127.0.0.1) and non-ipv4 addresses
                            return;
                        }

                        if (alias >= 1) {
                            // this single interface has multiple ipv4 addresses
                            //console.log(ifname + ':' + alias, iface.address);
                            ifacesarray.push({
                                name: ifname + ':' + alias,
                                address: iface.address
                            });
                        } else {
                            // this interface has only one ipv4 adress
                            //console.log(ifname, iface.address);
                            ifacesarray.push({
                                name: ifname,
                                address: iface.address
                            });
                        }
                        ++alias;
                    });
                });

                request({
                    url: 'http://bigdotsoftware.pl/inviteposnetserver.php',
                    method: 'POST',
                    rejectUnauthorized: false,
                    strictSSL: false,
                    json: {
                        'action': 'license',
                        'days': days,
                        'uniqueLicenseId': uniqueLicenseId,
                        'hostname': os.hostname(),
                        'platform': os.platform(),
                        'arch': os.arch(),
                        'uptime': os.uptime(),
                        'type': os.type(),
                        'ifacesarray': ifacesarray
                    }
                }, (err, internalres, body) => {
                    if (err)
                        logger.logger.info('cannot execute invitation: ' + err);
                    else
                        logger.logger.info('executed ok');
                });

                if (uniqueLicenseId != '' && days < 10) {
                    logger.logger.info('Obtaining new license file');
                    obtainNewLicenseFile(uniqueLicenseId);
                }

            }
        }
    }
    return stat;
}

module.exports.init = init;
